@@ -1,5 +1,3 @@
-#!/usr/bin/env python
-#
 # This file describes the common eFuses structures for chips
 #
 # SPDX-FileCopyrightText: 2020-2022 Espressif Systems (Shanghai) CO LTD
@@ -10,7 +8,7 @@ import binascii
 import re
 import sys
 
-from bitstring import BitArray, BitString, CreationError
+from bitstring import BitArray, BitStream, CreationError
 
 import esptool
 
@@ -155,14 +153,14 @@ class EfuseBlockBase(EfuseProtectBase):
         self.len = param.len
         self.key_purpose_name = param.key_purpose
         bit_block_len = self.get_block_len() * 8
-        self.bitarray = BitString(bit_block_len)
+        self.bitarray = BitStream(bit_block_len)
         self.bitarray.set(0)
-        self.wr_bitarray = BitString(bit_block_len)
+        self.wr_bitarray = BitStream(bit_block_len)
         self.wr_bitarray.set(0)
         self.fail = False
         self.num_errors = 0
         if self.id == 0:
-            self.err_bitarray = BitString(bit_block_len)
+            self.err_bitarray = BitStream(bit_block_len)
             self.err_bitarray.set(0)
         else:
             self.err_bitarray = None
@@ -336,7 +334,7 @@ class EfuseBlockBase(EfuseProtectBase):
         # in bitstring      = [N] ... [2][1][0]            (to get a correct bitstring
         #                                                   need to reverse new_data)
         # *[x] - means a byte.
-        data = BitString(bytes=new_data[::-1], length=len(new_data) * 8)
+        data = BitStream(bytes=new_data[::-1], length=len(new_data) * 8)
         if self.parent.debug:
             print(
                 "\twritten : {} ->\n\tto write: {}".format(self.get_bitstring(), data)
@@ -458,6 +456,13 @@ class EspEfusesBase(object):
     def efuse_controller_setup(self):
         pass
 
+    def reconnect_chip(self, esp):
+        print("Re-connecting...")
+        baudrate = esp._port.baudrate
+        port = esp._port.port
+        esp._port.close()
+        return esptool.cmds.detect_chip(port, baudrate)
+
     def get_index_block_by_name(self, name):
         for block in self.blocks:
             if block.name == name or name in block.alias:
@@ -556,7 +561,7 @@ class EfuseFieldBase(EfuseProtectBase):
             field_len = int(re.search(r"\d+", self.efuse_type).group())
             if self.efuse_type.startswith("bytes"):
                 field_len *= 8
-        self.bitarray = BitString(field_len)
+        self.bitarray = BitStream(field_len)
         self.bit_len = field_len
         self.bitarray.set(0)
         self.update(self.parent.blocks[self.block].bitarray)
@@ -653,6 +658,9 @@ class EfuseFieldBase(EfuseProtectBase):
         self.save_to_block(bitarray_field)
 
     def update(self, bit_array_block):
+        if self.word is None or self.pos is None:
+            self.bitarray.overwrite(self.convert_to_bitstring(self.get()), pos=0)
+            return
         field_len = self.bitarray.len
         bit_array_block.pos = bit_array_block.length - (
             self.word * 32 + self.pos + field_len
