@@ -1,6 +1,4 @@
-#!/usr/bin/env python
-#
-# SPDX-FileCopyrightText: 2016-2022 Espressif Systems (Shanghai) CO LTD
+# SPDX-FileCopyrightText: 2016-2023 Espressif Systems (Shanghai) CO LTD
 #
 # SPDX-License-Identifier: GPL-2.0-or-later
 
@@ -10,6 +8,7 @@ import operator
 import os
 import struct
 import sys
+import tempfile
 import zlib
 from collections import namedtuple
 
@@ -35,6 +34,11 @@ SIG_BLOCK_VERSION_ECDSA = 0x03
 CURVE_ID_P192 = 1
 CURVE_ID_P256 = 2
 
+SECTOR_SIZE = 4096
+SIG_BLOCK_SIZE = (
+    1216  # Refer to secure boot v2 signature block format for more details.
+)
+
 
 def get_chunks(source, chunk_len):
     """Returns an iterator over 'chunk_len' chunks of 'source'"""
@@ -202,11 +206,22 @@ def _load_ecdsa_signing_key(keyfile):
     if sk.curve != ecdsa.NIST256p:
         raise esptool.FatalError(
             "Signing key uses incorrect curve. ESP32 Secure Boot only supports "
-            "NIST256p (openssl calls this curve 'prime256v1"
+            "NIST256p (openssl calls this curve 'prime256v1')"
         )
     return sk
 
 
+def _load_ecdsa_verifying_key(keyfile):
+    """Load ECDSA verifying key for Secure Boot V1 only"""
+    vk = ecdsa.VerifyingKey.from_pem(keyfile.read())
+    if vk.curve != ecdsa.NIST256p:
+        raise esptool.FatalError(
+            "Signing key uses incorrect curve. ESP32 Secure Boot only supports "
+            "NIST256p (openssl calls this curve 'prime256v1')"
+        )
+    return vk
+
+
 def _load_sbv2_signing_key(keydata):
     """
     Load Secure Boot V2 signing key
@@ -263,7 +278,11 @@ def _load_sbv2_pub_key(keydata):
 
 def _get_sbv2_pub_key(keyfile):
     key_data = keyfile.read()
-    if b"-BEGIN RSA PRIVATE KEY" in key_data or b"-BEGIN EC PRIVATE KEY" in key_data:
+    if (
+        b"-BEGIN RSA PRIVATE KEY" in key_data
+        or b"-BEGIN EC PRIVATE KEY" in key_data
+        or b"-BEGIN PRIVATE KEY" in key_data
+    ):
         return _load_sbv2_signing_key(key_data).public_key()
     elif b"-BEGIN PUBLIC KEY" in key_data:
         vk = _load_sbv2_pub_key(key_data)
@@ -303,7 +322,8 @@ def _microecc_format(a, b, curve_len):
 
 
 def sign_data(args):
-    _check_output_is_not_input(args.keyfile, args.output)
+    if args.keyfile:
+        _check_output_is_not_input(args.keyfile, args.output)
     _check_output_is_not_input(args.datafile, args.output)
     if args.version == "1":
         return sign_secure_boot_v1(args)
@@ -315,18 +335,33 @@ def sign_secure_boot_v1(args):
     """
     Sign a data file with a ECDSA private key, append binary signature to file contents
     """
-    if len(args.keyfile) > 1:
-        raise esptool.FatalError("Secure Boot V1 only supports one signing key")
-    sk = _load_ecdsa_signing_key(args.keyfile[0])
-
-    # calculate signature of binary data
     binary_content = args.datafile.read()
-    signature = sk.sign_deterministic(binary_content, hashlib.sha256)
+
+    if args.hsm:
+        raise esptool.FatalError(
+            "Secure Boot V1 does not support signing using an "
+            "external Hardware Security Module (HSM)"
+        )
+
+    if args.signature:
+        print("Pre-calculated signatures found")
+        if len(args.pub_key) > 1:
+            raise esptool.FatalError("Secure Boot V1 only supports one signing key")
+        signature = args.signature[0].read()
+        # get verifying/public key
+        vk = _load_ecdsa_verifying_key(args.pub_key[0])
+    else:
+        if len(args.keyfile) > 1:
+            raise esptool.FatalError("Secure Boot V1 only supports one signing key")
+        sk = _load_ecdsa_signing_key(args.keyfile[0])
+
+        # calculate signature of binary data
+        signature = sk.sign_deterministic(binary_content, hashlib.sha256)
+        # get verifying/public key
+        vk = sk.get_verifying_key()
 
     # back-verify signature
-    vk = sk.get_verifying_key()
     vk.verify(signature, binary_content, hashlib.sha256)  # throws exception on failure
-
     if args.output is None or os.path.abspath(args.output) == os.path.abspath(
         args.datafile.name
     ):  # append signature to input file
@@ -340,10 +375,7 @@ def sign_secure_boot_v1(args):
     )  # Version indicator, allow for different curves/formats later
     outfile.write(signature)
     outfile.close()
-    print(
-        "Signed %d bytes of data from %s with key %s"
-        % (len(binary_content), args.datafile.name, args.keyfile[0].name)
-    )
+    print("Signed %d bytes of data from %s" % (len(binary_content), args.datafile.name))
 
 
 def sign_secure_boot_v2(args):
@@ -353,31 +385,30 @@ def sign_secure_boot_v2(args):
 
     Write output file with a Secure Boot V2 header appended.
     """
-    SECTOR_SIZE = 4096
-    SIG_BLOCK_SIZE = 1216
     SIG_BLOCK_MAX_COUNT = 3
-
-    signature_sector = b""
-    key_count = len(args.keyfile)
     contents = args.datafile.read()
+    sig_block_num = 0
+    signature_sector = b""
 
-    if key_count > SIG_BLOCK_MAX_COUNT:
-        print(
-            "WARNING: Upto %d signing keys are supported for ESP32-S2. "
-            "For ESP32-ECO3 only 1 signing key is supported",
-            SIG_BLOCK_MAX_COUNT,
-        )
+    signature = args.signature
+    pub_key = args.pub_key
 
     if len(contents) % SECTOR_SIZE != 0:
-        pad_by = SECTOR_SIZE - (len(contents) % SECTOR_SIZE)
-        print(
-            "Padding data contents by %d bytes "
-            "so signature sector aligns at sector boundary" % pad_by
-        )
-        contents += b"\xff" * pad_by
-    elif args.append_signatures:
-        sig_block_num = 0
+        if args.signature:
+            raise esptool.FatalError(
+                "Secure Boot V2 requires the signature block to start "
+                "from a 4KB aligned sector "
+                "but the datafile supplied is not sector aligned."
+            )
+        else:
+            pad_by = SECTOR_SIZE - (len(contents) % SECTOR_SIZE)
+            print(
+                f"Padding data contents by {pad_by} bytes "
+                "so signature sector aligns at sector boundary"
+            )
+            contents += b"\xff" * pad_by
 
+    elif args.append_signatures:
         while sig_block_num < SIG_BLOCK_MAX_COUNT:
             sig_block = validate_signature_block(contents, sig_block_num)
             if sig_block is None:
@@ -387,7 +418,8 @@ def sign_secure_boot_v2(args):
             )
             sig_block_num += 1
 
-        assert len(signature_sector) % SIG_BLOCK_SIZE == 0
+        if len(signature_sector) % SIG_BLOCK_SIZE != 0:
+            raise esptool.FatalError("Incorrect signature sector size")
 
         if sig_block_num == 0:
             print(
@@ -396,27 +428,179 @@ def sign_secure_boot_v2(args):
             )
         else:
             print(
-                "%d valid signature block(s) already present in the signature sector."
-                % sig_block_num
+                f"{sig_block_num} valid signature block(s) already present "
+                "in the signature sector."
             )
-
-            empty_signature_blocks = SIG_BLOCK_MAX_COUNT - sig_block_num
-            if key_count > empty_signature_blocks:
+            if sig_block_num == SIG_BLOCK_MAX_COUNT:
                 raise esptool.FatalError(
-                    "Number of keys(%d) more than the empty signature blocks.(%d)"
-                    % (key_count, empty_signature_blocks)
+                    f"Upto {SIG_BLOCK_MAX_COUNT} signature blocks are supported. "
+                    "(For ESP32-ECO3 only 1 signature block is supported)"
                 )
+
             # Signature stripped off the content
             # (the legitimate blocks are included in signature_sector)
             contents = contents[: len(contents) - SECTOR_SIZE]
 
-    print("%d signing key(s) found." % key_count)
+    if args.hsm:
+        if args.hsm_config is None:
+            raise esptool.FatalError(
+                "Config file is required to generate signature using an external HSM."
+            )
+        import espsecure.esp_hsm_sign as hsm
+
+        try:
+            config = hsm.read_hsm_config(args.hsm_config)
+        except Exception as e:
+            raise esptool.FatalError(f"Incorrect HSM config file format ({e})")
+        if pub_key is None:
+            pub_key = extract_pubkey_from_hsm(config)
+        signature = generate_signature_using_hsm(config, contents)
+
+    if signature:
+        print("Pre-calculated signatures found")
+        key_count = len(pub_key)
+        if len(signature) != key_count:
+            raise esptool.FatalError(
+                f"Number of public keys ({key_count}) not equal to "
+                f"the number of signatures {len(signature)}."
+            )
+    else:
+        key_count = len(args.keyfile)
+
+    empty_signature_blocks = SIG_BLOCK_MAX_COUNT - sig_block_num
+    if key_count > empty_signature_blocks:
+        raise esptool.FatalError(
+            f"Number of keys({key_count}) more than the empty signature blocks."
+            f"({empty_signature_blocks})"
+        )
+
+    print(f"{key_count} signing key(s) found.")
     # Calculate digest of data file
     digest = hashlib.sha256()
     digest.update(contents)
     digest = digest.digest()
 
-    for keyfile in args.keyfile:
+    # Generate signature block using pre-calculated signatures
+    if signature:
+        signature_block = generate_signature_block_using_pre_calculated_signature(
+            signature, pub_key, digest
+        )
+    # Generate signature block by signing using private keys
+    else:
+        signature_block = generate_signature_block_using_private_key(
+            args.keyfile, digest
+        )
+
+    if signature_block is None or len(signature_block) == 0:
+        raise esptool.FatalError("Signature Block generation failed")
+
+    signature_sector += signature_block
+
+    if (
+        len(signature_sector) < 0
+        and len(signature_sector) > SIG_BLOCK_SIZE * 3
+        and len(signature_sector) % SIG_BLOCK_SIZE != 0
+    ):
+        raise esptool.FatalError("Incorrect signature sector generation")
+
+    total_sig_blocks = len(signature_sector) // SIG_BLOCK_SIZE
+
+    # Pad signature_sector to sector
+    signature_sector = signature_sector + (
+        b"\xff" * (SECTOR_SIZE - len(signature_sector))
+    )
+    if len(signature_sector) != SECTOR_SIZE:
+        raise esptool.FatalError("Incorrect signature sector size")
+
+    # Write to output file, or append to existing file
+    if args.output is None:
+        args.datafile.close()
+        args.output = args.datafile.name
+    with open(args.output, "wb") as f:
+        f.write(contents + signature_sector)
+    print(
+        f"Signed {len(contents)} bytes of data from {args.datafile.name}. "
+        f"Signature sector now has {total_sig_blocks} signature blocks."
+    )
+
+
+def generate_signature_using_hsm(config, contents):
+    import espsecure.esp_hsm_sign as hsm
+
+    session = hsm.establish_session(config)
+    # get the private key
+    private_key = hsm.get_privkey_info(session, config)
+    # Sign payload
+    signature = hsm.sign_payload(private_key, contents)
+    hsm.close_connection(session)
+    temp_signature_file = tempfile.TemporaryFile()
+    temp_signature_file.write(signature)
+    temp_signature_file.seek(0)
+    return [temp_signature_file]
+
+
+def generate_signature_block_using_pre_calculated_signature(signature, pub_key, digest):
+    signature_blocks = b""
+    for sig, pk in zip(signature, pub_key):
+        try:
+            public_key = _get_sbv2_pub_key(pk)
+            signature = sig.read()
+            if isinstance(public_key, rsa.RSAPublicKey):
+                # RSA signature
+                rsa_primitives = _get_sbv2_rsa_primitives(public_key)
+                # Verify the signature
+                public_key.verify(
+                    signature,
+                    digest,
+                    padding.PSS(mgf=padding.MGF1(hashes.SHA256()), salt_length=32),
+                    utils.Prehashed(hashes.SHA256()),
+                )
+
+                signature_block = generate_rsa_signature_block(
+                    digest, rsa_primitives, signature
+                )
+            else:
+                # ECDSA signature
+                numbers = public_key.public_numbers()
+                if isinstance(numbers.curve, ec.SECP192R1):
+                    curve_len = 192
+                    curve_id = CURVE_ID_P192
+                elif isinstance(numbers.curve, ec.SECP256R1):
+                    curve_len = 256
+                    curve_id = CURVE_ID_P256
+                else:
+                    raise esptool.FatalError("Invalid ECDSA curve instance.")
+
+                # Verify the signature
+                public_key.verify(
+                    signature, digest, ec.ECDSA(utils.Prehashed(hashes.SHA256()))
+                )
+
+                pubkey_point = _microecc_format(numbers.x, numbers.y, curve_len)
+                r, s = utils.decode_dss_signature(signature)
+                signature_rs = _microecc_format(r, s, curve_len)
+                signature_block = generate_ecdsa_signature_block(
+                    digest, curve_id, pubkey_point, signature_rs
+                )
+        except exceptions.InvalidSignature:
+            raise esptool.FatalError(
+                "Signature verification failed: Invalid Signature\n"
+                "The pre-calculated signature has not been signed "
+                "using the given public key"
+            )
+        signature_block += struct.pack("<I", zlib.crc32(signature_block) & 0xFFFFFFFF)
+        signature_block += b"\x00" * 16  # padding
+
+        if len(signature_block) != SIG_BLOCK_SIZE:
+            raise esptool.FatalError("Incorrect signature block size")
+
+        signature_blocks += signature_block
+    return signature_blocks
+
+
+def generate_signature_block_using_private_key(keyfiles, digest):
+    signature_blocks = b""
+    for keyfile in keyfiles:
         private_key = _load_sbv2_signing_key(keyfile.read())
 
         # Sign
@@ -431,23 +615,8 @@ def sign_secure_boot_v2(args):
                 utils.Prehashed(hashes.SHA256()),
             )
             rsa_primitives = _get_sbv2_rsa_primitives(private_key.public_key())
-
-            # Encode in signature block format
-            #
-            # Note: the [::-1] is to byte swap all of the bignum
-            # values (signatures, coefficients) to little endian
-            # for use with the RSA peripheral, rather than big endian
-            # which is conventionally used for RSA.
-            signature_block = struct.pack(
-                "<BBxx32s384sI384sI384s",
-                SIG_BLOCK_MAGIC,
-                SIG_BLOCK_VERSION_RSA,
-                digest,
-                int_to_bytes(rsa_primitives.n)[::-1],
-                rsa_primitives.e,
-                int_to_bytes(rsa_primitives.rinv)[::-1],
-                rsa_primitives.m & 0xFFFFFFFF,
-                signature[::-1],
+            signature_block = generate_rsa_signature_block(
+                digest, rsa_primitives, signature
             )
         else:
             # ECDSA signature
@@ -459,56 +628,70 @@ def sign_secure_boot_v2(args):
             if isinstance(private_key.curve, ec.SECP192R1):
                 curve_len = 192
                 curve_id = CURVE_ID_P192
-            else:
+            elif isinstance(numbers.curve, ec.SECP256R1):
                 curve_len = 256
                 curve_id = CURVE_ID_P256
+            else:
+                raise esptool.FatalError("Invalid ECDSA curve instance.")
 
             pubkey_point = _microecc_format(numbers.x, numbers.y, curve_len)
 
             r, s = utils.decode_dss_signature(signature)
             signature_rs = _microecc_format(r, s, curve_len)
-
-            # block is padded out to the much larger size
-            # of the RSA version of this structure
-            signature_block = struct.pack(
-                "<BBxx32sB64s64s1031x",
-                SIG_BLOCK_MAGIC,
-                SIG_BLOCK_VERSION_ECDSA,
-                digest,
-                curve_id,
-                pubkey_point,
-                signature_rs,
+            signature_block = generate_ecdsa_signature_block(
+                digest, curve_id, pubkey_point, signature_rs
             )
 
         signature_block += struct.pack("<I", zlib.crc32(signature_block) & 0xFFFFFFFF)
         signature_block += b"\x00" * 16  # padding
 
-        assert len(signature_block) == SIG_BLOCK_SIZE
-        signature_sector += signature_block
+        if len(signature_block) != SIG_BLOCK_SIZE:
+            raise esptool.FatalError("Incorrect signature block size")
 
-    assert (
-        len(signature_sector) > 0
-        and len(signature_sector) <= SIG_BLOCK_SIZE * 3
-        and len(signature_sector) % SIG_BLOCK_SIZE == 0
-    )
-    total_sig_blocks = len(signature_sector) // SIG_BLOCK_SIZE
+        signature_blocks += signature_block
+    return signature_blocks
 
-    # Pad signature_sector to sector
-    signature_sector = signature_sector + (
-        b"\xff" * (SECTOR_SIZE - len(signature_sector))
-    )
-    assert len(signature_sector) == SECTOR_SIZE
 
-    # Write to output file, or append to existing file
-    if args.output is None:
-        args.datafile.close()
-        args.output = args.datafile.name
-    with open(args.output, "wb") as f:
-        f.write(contents + signature_sector)
-    print(
-        "Signed %d bytes of data from %s. Signature sector now has %d signature blocks."
-        % (len(contents), args.datafile.name, total_sig_blocks)
+def generate_rsa_signature_block(digest, rsa_primitives, signature):
+    """
+    Encode in rsa signature block format
+
+    Note: the [::-1] is to byte swap all of the bignum
+    values (signatures, coefficients) to little endian
+    for use with the RSA peripheral, rather than big endian
+    which is conventionally used for RSA.
+    """
+    signature_block = struct.pack(
+        "<BBxx32s384sI384sI384s",
+        SIG_BLOCK_MAGIC,
+        SIG_BLOCK_VERSION_RSA,
+        digest,
+        int_to_bytes(rsa_primitives.n)[::-1],
+        rsa_primitives.e,
+        int_to_bytes(rsa_primitives.rinv)[::-1],
+        rsa_primitives.m & 0xFFFFFFFF,
+        signature[::-1],
     )
+    return signature_block
+
+
+def generate_ecdsa_signature_block(digest, curve_id, pubkey_point, signature_rs):
+    """
+    Encode in rsa signature block format
+
+    # block is padded out to the much larger size
+    # of the RSA version of this structure
+    """
+    signature_block = struct.pack(
+        "<BBxx32sB64s64s1031x",
+        SIG_BLOCK_MAGIC,
+        SIG_BLOCK_VERSION_ECDSA,
+        digest,
+        curve_id,
+        pubkey_point,
+        signature_rs,
+    )
+    return signature_block
 
 
 def verify_signature(args):
@@ -559,11 +742,6 @@ def verify_signature_v1(args):
 
 
 def validate_signature_block(image_content, sig_blk_num):
-    SECTOR_SIZE = 4096
-    SIG_BLOCK_SIZE = (
-        1216  # Refer to secure boot v2 signature block format for more details.
-    )
-
     offset = -SECTOR_SIZE + sig_blk_num * SIG_BLOCK_SIZE
     sig_blk = image_content[offset : offset + SIG_BLOCK_SIZE]
     assert len(sig_blk) == SIG_BLOCK_SIZE
@@ -583,19 +761,30 @@ def validate_signature_block(image_content, sig_blk_num):
 
     if is_invalid_block or blk_crc != calc_crc & 0xFFFFFFFF:  # Signature block invalid
         return None
-
-    print(
-        "Signature block %d is valid (%s). "
-        % (sig_blk_num, "RSA" if version == SIG_BLOCK_VERSION_RSA else "ECDSA")
-    )
+    key_type = "RSA" if version == SIG_BLOCK_VERSION_RSA else "ECDSA"
+    print(f"Signature block {sig_blk_num} is valid ({key_type}).")
     return sig_blk
 
 
 def verify_signature_v2(args):
     """Verify a previously signed binary image, using the RSA or ECDSA public key"""
-    SECTOR_SIZE = 4096
 
-    vk = _get_sbv2_pub_key(args.keyfile)
+    keyfile = args.keyfile
+    if args.hsm:
+        if args.hsm_config is None:
+            raise esptool.FatalError(
+                "Config file is required to extract public key from an external HSM."
+            )
+        import espsecure.esp_hsm_sign as hsm
+
+        try:
+            config = hsm.read_hsm_config(args.hsm_config)
+        except Exception as e:
+            raise esptool.FatalError(f"Incorrect HSM config file format ({e})")
+        # get public key from HSM
+        keyfile = extract_pubkey_from_hsm(config)[0]
+
+    vk = _get_sbv2_pub_key(keyfile)
 
     if isinstance(vk, rsa.RSAPublicKey):
         SIG_BLOCK_MAX_COUNT = 3
@@ -617,14 +806,14 @@ def verify_signature_v2(args):
     for sig_blk_num in range(SIG_BLOCK_MAX_COUNT):
         sig_blk = validate_signature_block(image_content, sig_blk_num)
         if sig_blk is None:
-            print("Signature block %d invalid. Skipping." % sig_blk_num)
+            print(f"Signature block {sig_blk_num} invalid. Skipping.")
             continue
         _, version, blk_digest = struct.unpack("<BBxx32s", sig_blk[:36])
 
         if blk_digest != digest:
             raise esptool.FatalError(
                 "Signature block image digest does not match "
-                "the actual image digest %s. Expected %s." % (digest, blk_digest)
+                f"the actual image digest {digest}. Expected {blk_digest}."
             )
 
         try:
@@ -655,20 +844,19 @@ def verify_signature_v2(args):
                 signature = utils.encode_dss_signature(r, s)
 
                 vk.verify(signature, digest, ec.ECDSA(utils.Prehashed(hashes.SHA256())))
+
+            key_type = "RSA" if isinstance(vk, rsa.RSAPublicKey) else "ECDSA"
+
             print(
-                "Signature block %d verification successful with %s (%s)."
-                % (
-                    sig_blk_num,
-                    args.keyfile.name,
-                    "RSA" if isinstance(vk, rsa.RSAPublicKey) else "ECDSA",
-                )
+                f"Signature block {sig_blk_num} verification successful using "
+                f"the supplied key ({key_type})."
             )
             valid = True
 
         except exceptions.InvalidSignature:
             print(
-                "Signature block %d is not signed by %s. Checking the next block"
-                % (sig_blk_num, args.keyfile.name)
+                f"Signature block {sig_blk_num} is not signed by the supplied key. "
+                "Checking the next block"
             )
             continue
 
@@ -704,6 +892,24 @@ def extract_public_key(args):
     )
 
 
+def extract_pubkey_from_hsm(config):
+    import espsecure.esp_hsm_sign as hsm
+
+    session = hsm.establish_session(config)
+    # get public key from HSM
+    public_key = hsm.get_pubkey(session, config)
+    hsm.close_connection(session)
+
+    pem = public_key.public_bytes(
+        encoding=serialization.Encoding.PEM,
+        format=serialization.PublicFormat.SubjectPublicKeyInfo,
+    )
+    temp_pub_key_file = tempfile.TemporaryFile()
+    temp_pub_key_file.write(pem)
+    temp_pub_key_file.seek(0)
+    return [temp_pub_key_file]
+
+
 def _sha256_digest(data):
     digest = hashlib.sha256()
     digest.update(data)
@@ -715,11 +921,7 @@ def signature_info_v2(args):
     Validates the signature block and prints the RSA/ECDSA public key
     digest for valid blocks
     """
-    SECTOR_SIZE = 4096
     SIG_BLOCK_MAX_COUNT = 3
-    SIG_BLOCK_SIZE = (
-        1216  # Refer to secure boot v2 signature block format for more details.
-    )
 
     image_content = args.datafile.read()
     if len(image_content) < SECTOR_SIZE or len(image_content) % SECTOR_SIZE != 0:
@@ -1050,7 +1252,7 @@ def _flash_encryption_operation_aes_xts(
 
     inblocks = _split_blocks(indata, 0x80)  # split into 1024 bit blocks
 
-    output = b""
+    output = []
     for inblock in inblocks:  # for each block
         tweak = struct.pack("<I", (flash_address & ~0x7F)) + (b"\x00" * 12)
         flash_address += 0x80  # for next block
@@ -1065,7 +1267,9 @@ def _flash_encryption_operation_aes_xts(
 
         inblock = inblock[::-1]  # reverse input
         outblock = encryptor.update(inblock)  # standard algo
-        output += outblock[::-1]  # reverse output
+        output.append(outblock[::-1])  # reverse output
+
+    output = b"".join(output)
 
     # undo any padding we applied to the input
     if pad_right != 0:
@@ -1087,9 +1291,10 @@ def _flash_encryption_operation_aes_xts(
 def _split_blocks(text, block_len=16):
     """Take a bitstring, split it into chunks of "block_len" each"""
     assert len(text) % block_len == 0
-    while len(text) > 0:
-        yield text[0:block_len]
-        text = text[block_len:]
+    pos = 0
+    while pos < len(text):
+        yield text[pos : pos + block_len]
+        pos = pos + block_len
 
 
 def decrypt_flash_data(args):
@@ -1277,16 +1482,42 @@ def main(custom_commandline=None):
         "-k",
         help="Private key file for signing. Key is in PEM format.",
         type=argparse.FileType("rb"),
-        required=True,
         nargs="+",
     )
     p.add_argument(
         "--append_signatures",
         "-a",
-        help="Append signature block(s) to already signed image"
+        help="Append signature block(s) to already signed image. "
         "Valid only for ESP32-S2.",
         action="store_true",
     )
+    p.add_argument(
+        "--hsm",
+        help="Use an external Hardware Security Module "
+        "to generate signature using PKCS#11 interface.",
+        action="store_true",
+    )
+    p.add_argument(
+        "--hsm-config",
+        help="Config file for the external Hardware Security Module "
+        "to be used to generate signature.",
+        default=None,
+    )
+    p.add_argument(
+        "--pub-key",
+        help="Public key files corresponding to the private key used to generate "
+        "the pre-calculated signatures. Keys should be in PEM format.",
+        type=argparse.FileType("rb"),
+        nargs="+",
+    )
+    p.add_argument(
+        "--signature",
+        help="Pre-calculated signatures. "
+        "Signatures generated using external private keys e.g. keys stored in HSM.",
+        type=argparse.FileType("rb"),
+        nargs="+",
+        default=None,
+    )
     p.add_argument(
         "--output",
         "-o",
@@ -1311,13 +1542,24 @@ def main(custom_commandline=None):
         choices=["1", "2"],
         required=True,
     )
+    p.add_argument(
+        "--hsm",
+        help="Use an external Hardware Security Module "
+        "to verify signature using PKCS#11 interface.",
+        action="store_true",
+    )
+    p.add_argument(
+        "--hsm-config",
+        help="Config file for the external Hardware Security Module "
+        "to be used to verify signature.",
+        default=None,
+    )
     p.add_argument(
         "--keyfile",
         "-k",
         help="Public key file for verification. "
         "Can be private or public key in PEM format.",
         type=argparse.FileType("rb"),
-        required=True,
     )
     p.add_argument(
         "datafile",
@@ -1448,7 +1690,8 @@ def main(custom_commandline=None):
     p.add_argument(
         "--aes_xts",
         "-x",
-        help="Decrypt data using AES-XTS as used on ESP32-S2, ESP32-C2 and ESP32-C3",
+        help="Decrypt data using AES-XTS as used on "
+        "ESP32-S2, ESP32-C2, ESP32-C3 and ESP32-C6",
         action="store_true",
     )
     p.add_argument(
@@ -1487,7 +1730,8 @@ def main(custom_commandline=None):
     p.add_argument(
         "--aes_xts",
         "-x",
-        help="Encrypt data using AES-XTS as used on ESP32-S2, ESP32-C2 and ESP32-C3",
+        help="Encrypt data using AES-XTS as used on "
+        "ESP32-S2, ESP32-C2, ESP32-C3 and ESP32-C6",
         action="store_true",
     )
     p.add_argument(
